#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <sstream>

/*
  Brainfuck to Shakespear Programming Language converter.
  Written by Jonas Sjbergh, jsh@nada.kth.se

  Puck's stack is memory to the left (below) of 'p',
  Ajax's stack is memory to the right of 'p',
  Page is '*p',
  Page and Puck is always on stage when a Scene ends.
*/

/*
  Convert int i to a string with 
  i written with roman numerals.
*/
std::string to_roman(int i) {
  std::string s = "";
  while(i >= 1000) {
    s += "M";
    i -= 1000;
  }
  if(i >= 900) {
    s += "CM";
    i -= 900;
  }
  if(i >= 500) {
    s += "D";
    i -= 500;
  }
  if(i >= 400) {
    s += "CD";
    i -= 400;
  }
  while(i >= 100) {
    s += "C";
    i -= 100;
  }
  if(i >= 90) {
    s += "XC";
    i -= 90;
  }
  if(i >= 50) {
    s += "L";
    i -= 50;
  }
  if(i >= 40) {
    s += "XL";
    i -= 40;
  }
  while(i >= 10) {
    s += "X";
    i -= 10;
  }
  if(i >= 9) {
    s += "IX";
    i -= 9;
  }
  if(i >= 5) {
    s += "V";
    i -= 5;
  }
  if(i == 4) {
    s += "IV";
    i -= 4;
  }
  while(i > 0) {
    s += "I";
    i -= 1;
  }
  return s;
}

/*
  Optimize output of strings of "+" or "-" a little.
  This can be done better, but this is at least
  better than "n" lines of "add one".
*/
void do_plusminus(int n, std::ostream &os, const char *cat_pig) {
  std::string s;
  for(; n > 0; n /= 2) {
    if(n & 1)
      os << "Puck: You are as bad as the sum of thyself and a " 
	 << s << cat_pig << "." << std::endl;
    s += "red ";
  }
}

int main(int argc, char **argv) {
  bool add_dummy = false;
  /* How large memory vector do we need?
     Poor Ajax will have this many zeros
     pushed onto his stack before the actual
     program starts */
  int memsize = 30000;
  if(argc > 1) {
    istringstream i(argv[1]);
    i >> memsize;
    if(memsize < 1)
      memsize = 100;
  }

  /* Give user something to read while running */
  std::cerr << "Running Brainfuck to Shakespear Programming Language converter"
	    << std::endl << "Usage: " << argv[0] << " [memsize] < BF-file > SPL-file" << std::endl
	    << "where memsize is the size of the Brainfuck memory vector to use."
	    << std::endl << "Using Brainfuck memory size " << memsize << std::endl
	    << "Preprocessing file, finding labels." << std::endl;
  std::vector<int> stacken;
  std::map<int, int> m;
  const int StartScene = 4;
  int mscene = StartScene;
  

  /* 
     Pass over bf-code once to get all jump-to labels, save
     bf-code in memory for another pass.
  */
  ostringstream os;  
  for(char c = std::cin.get(); !std::cin.eof(); c = std::cin.get()) {
    int temp, temp2;

    switch(c) {
    case '[':
      ++mscene;
      stacken.push_back(mscene);
      os << c;
      break;
    case ']':
      temp = stacken.back(); stacken.pop_back();
      temp2 = ++mscene;
      m[temp] = temp2;
    case '.':
    case ',':
    case '<':
    case '>':
    case '+':
    case '-':
      os << c;
      break;
    default:
      ;
    }
  }

  /* 
     Do another pass over the bf-code, this time output
     spl-code.
  */
  std::cerr << std::endl << "Preprocessing done. Generating Code." << std::endl;
  mscene = StartScene;

  /* write header */
  std::cout << "Brainfuck to SPL made this file, bf2spl made by Jonas Sjobergh"
	    << ", jsh@nada<dot>kth<dot>se ." 
	    << std::endl << "Puck, low memory."
	    << std::endl << "Ajax, high memory."
	    << std::endl << "Page, current."
	    << std::endl << "Act I: Where all code is."
	    << std::endl << "Scene I: Initialize Page to 0 and Puck to memsize." << std::endl
	    << "[Enter Ajax and Page]" << std::endl
	    << "Ajax: You are nothing!" << std::endl  
	    << "[Exit Page]\n[Enter Puck]\nAjax: ";

  /* fill memory with 0s */
  std::string s = "";
  for( ; memsize > 0; memsize /= 2) {
    if(memsize & 1) 
      std::cout << "You are as bad as the sum of thyself and a" << s << " cat."<< std::endl;
    s += " red";
  }
  std::cout << "Scene II: initialize memory vector." << std::endl
	    << "Puck: Am I better than zero?"
	    << std::endl << "If so, let us proceed to Scene III." << std::endl
	    << "Let us proceed to Scene IV." << std::endl
	    << "Scene III: Push Page onto Ajax." << std::endl
	    << "Ajax: You are as bad as the difference between thyself and a cat."
	    << std::endl << "[Exit Puck]\n[Enter Page]" << std::endl
	    << "Page: Remember me." << std::endl 
	    << "[Exit Page]\n[Enter Puck]\nPuck: Let us return to Scene II."
	    << std::endl << "Scene IV: The program starts here." << std::endl
	    << "[Exit Ajax]\n[Enter Page]" << std::endl;

  /* output actual bf-code (in spl) */
  int noofplus = 0;
  int noofminus = 0;
  istringstream is(os.str());  
  for(char c = is.get(); !is.eof(); c = is.get()) {
    std::cerr << c; // progress report for user, handled instructions
    int temp, temp2;

    if(noofminus && c != '-') {
      do_plusminus(noofminus, std::cout, "pig");
      noofminus = 0;
    }
    if(noofplus && c != '+') {
      do_plusminus(noofplus, std::cout, "cat");
      noofplus = 0;
    }

    switch(c) {
    case '.':
      add_dummy = false;
      std::cout << "Puck: Speak your mind!" << std::endl;
      break;
    case ',':
      add_dummy = false;
      std::cout << "Puck: Open your mind!" << std::endl;
      break;
    case '<':
      add_dummy = false;
      std::cout << "[Exit Puck]\n[Enter Ajax]\nPage: Remember me." << std::endl;      
      std::cout << "[Exit Ajax]\n[Enter Puck]\nPage: Recall s!\nPuck: You are as bad as me." << std::endl;
      break;
    case '>':
      add_dummy = false;
      std::cout << "Page: Remember me." << std::endl;      
      std::cout << "[Exit Puck]\n[Enter Ajax]\nPage: Recall s!\nAjax: You are as bad as me.\n[Exit Ajax]\n[Enter Puck]" << std::endl;
      break;
    case '[':
      if(add_dummy)
	std::cout << "Page: You are as bad as me." << std::endl;
      add_dummy = false;
      std::cout << "Scene " << to_roman(++mscene) << ": " << mscene << "." << std::endl;
      std::cout << "Page: Am I as bad as zero?\nIf so, let us proceed to Scene " 
		<< to_roman(m[mscene]) << "." << std::endl;
      stacken.push_back(mscene);
      break;
    case ']':
      temp = stacken.back(); stacken.pop_back();
      std::cout << "Page: Let us return to Scene " << to_roman(temp) << "." << std::endl;
      temp2 = ++mscene;
      m[temp] = temp2;
      std::cout << "Scene " << to_roman(temp2) << ": from Scene " << temp << "." << std::endl;
      add_dummy = true; // add dummy statement if new Scene next, empty Scenes not allowed
      // this could/should be changed to not use a new Scene if first Scene is empty...
      break;
    case '+':
      add_dummy = false;
      ++noofplus;
      break;
    case '-':
      add_dummy = false;
      ++noofminus;      
      break;
    default:
      ;
    }
  }

  /* 
     Print out any non-handled "+" or "-", this
     will not affect the program output, but if 
     the source has it, the spl-code should have it.
  */
  do_plusminus(noofminus, std::cout, "pig");
  do_plusminus(noofplus, std::cout, "cat");
  /* 
     Add one dummy statement that does nothing,
     in case source ends with "]" (all Scenes
     must have some statements in SPL).
  */
  if(add_dummy)
    std::cout << "Page: You are as bad as me." << std::endl;
  std::cerr << std::endl;
  return 0;
}
